/**************************************************************************
   lexmark.c - SANE library for Lexmark scanners.
   Copyright (C) 2003-2005 Lexmark International, Inc.
   
   This file is part of the SANE package.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.

   As a special exception, the authors of SANE give permission for
   additional uses of the libraries contained in this release of SANE.

   The exception is that, if you link a SANE library with other files
   to produce an executable, this does not by itself cause the
   resulting executable to be covered by the GNU General Public
   License.  Your use of that executable is in no way restricted on
   account of linking the SANE library code into it.

   This exception does not, however, invalidate any other reasons why
   the executable file might be covered by the GNU General Public
   License.

   If you submit changes to SANE to the maintainers to be included in
   a subsequent release, you agree by submitting the changes that
   those changes may be distributed with this exception intact.

   If you write modifications of your own for SANE, it is your choice
   whether to permit this exception to apply to your modifications.
   If you do not wish that, delete this exception notice.  
   **************************************************************************/

#include "../include/sane/config.h"

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include "../include/sane/sane.h"
#include "../include/sane/sanei.h"
#include "../include/sane/saneopts.h"
#include "../include/sane/sanei_config.h"

#define BACKEND_NAME	lexmark
#include "../include/sane/sanei_backend.h"

#include "lexmark.h"
#include "../lexmark/llpddk.h"

//error codes
#define IO_READ_ERROR	-1
#define ADF_PAPER_JAM	-2

#define LEXMARK_CONFIG_FILE "lexmark.conf"
#define BUILD 0
#define LXK_MAX_OPTION_STRING_SIZE 255
#define MM_PER_INCH 25.4
#define SCANNER_IOCTL_PRODUCT _IOR('U', 0x21, int)

static Lexmark_Device * first_lexmark_device = 0;
static SANE_Int num_lexmark_device = 0;
static SANE_Device ** sane_device_list = 0;
static SANE_Bool initialized = SANE_FALSE;
static int eof = 0;
SANE_Int dpi, tl_x, tl_y, bt_x, bt_y;;

static long data_size = 0, data_ctr = 0;
static SANE_Bool device_cancelled = SANE_FALSE;
static SANE_Bool gui_cancelled = SANE_FALSE;
static int cancel_ctr = 0;
static SANE_Bool paperOnADF = SANE_FALSE;
static SANE_Bool paperJamOnADF = SANE_FALSE;
static SANE_Bool startADF = SANE_FALSE;
static SANE_Bool forcedADF = SANE_FALSE;
static SANE_Bool usbDetached = SANE_FALSE;
static SANE_Int oldDPI = 150;

static SANE_String_Const mode_list[] = {
  "Color", "Gray", 0};

static SANE_String_Const halftone_list[] = {
  "Error Diffusion", "Threshold", NULL };

static SANE_String_Const scan_type_list[] = {
  "Photo", "Text", "Mixed", NULL };  

static SANE_String_Const scansource_list[] = {
  "Flatbed", "Automatic Document Feeder", NULL };
  
static SANE_Int dpi_list[] = {
  9, 75, 150, 300, 600, 
  1200, 2400, 4800, 9600, 19200
  };

static SANE_Range scan_width_range = {
  0, 216, 1 };

static SANE_Range scan_height_range = {
  0, 297, 1 };

static SANE_Range signed_char_range = {
  -125, 126, 1 };

static SANE_Range unsigned_char_range = {
  0, 255, 1 };
  
static SANE_Range gamma_range = {
  0, 10, 1};

  
void getADFStatus(void)
{
  unsigned char adfStatus;
  llpddk_get_adf_status(&adfStatus);

  if(adfStatus == 0x00)
  {
 	  paperOnADF = SANE_FALSE;
	  paperJamOnADF = SANE_FALSE;
  }
  else if(adfStatus == 0x01)
  {
 	paperOnADF = SANE_TRUE;
	paperJamOnADF = SANE_FALSE;
  }
  else if(adfStatus == 0x0F)
  {
 	paperOnADF = SANE_TRUE;
	paperJamOnADF = SANE_TRUE;
  }
  else  //unknown response, reset to false
  {
    paperOnADF = SANE_FALSE;
    paperJamOnADF = SANE_FALSE;
  }
}

SANE_Status 
init_options(Lexmark_Device * lexmark_device)
 {

  SANE_Option_Descriptor * od;


  DBG (2, "init_options: lexmark_device = %p\n", lexmark_device);

  /* number of options */
  od = &(lexmark_device->opt[LXK_OPT_NUM_OPTS]);
  od->name = SANE_NAME_NUM_OPTIONS;
  od->title = SANE_TITLE_NUM_OPTIONS;
  od->desc = SANE_DESC_NUM_OPTIONS;
  od->type = SANE_TYPE_INT;
  od->unit = SANE_UNIT_NONE;
  od->size = sizeof (SANE_Word);
  od->cap = SANE_CAP_SOFT_DETECT;
  od->constraint_type = SANE_CONSTRAINT_NONE;
  od->constraint.range = 0;
  lexmark_device->val[LXK_OPT_NUM_OPTS].w = LXK_NUM_OPTIONS;
    
  /* channels */
  od = &(lexmark_device->opt[LXK_OPT_CHANNELS]);
  od->name = /*SANE_NAME_SCAN_MODE*/"Scan Mode";
  od->title = /*SANE_TITLE_SCAN_MODE*/"Scan Mode";
  od->desc = /*SANE_DESC_SCAN_MODE*/"Scan Mode";
  od->type = SANE_TYPE_STRING;
  od->unit = SANE_UNIT_NONE;
  od->size = LXK_MAX_OPTION_STRING_SIZE;
  od->cap = SANE_CAP_SOFT_DETECT | SANE_CAP_SOFT_SELECT;
  od->constraint_type = SANE_CONSTRAINT_STRING_LIST;
  od->constraint.string_list = mode_list;
  lexmark_device->val[LXK_OPT_CHANNELS].s = malloc(od->size);
  if (!lexmark_device->val[LXK_OPT_CHANNELS].s)
    return SANE_STATUS_NO_MEM;
  strcpy(lexmark_device->val[LXK_OPT_CHANNELS].s,"Color");
  
  /* halftone mode */
  od = &(lexmark_device->opt[LXK_OPT_HALFTONE_MODE]);
  od->name = SANE_NAME_HALFTONE;
  od->title = SANE_TITLE_HALFTONE;
  od->desc = SANE_DESC_HALFTONE;
  od->type = SANE_TYPE_STRING;
  od->unit = SANE_UNIT_NONE;
  od->size = LXK_MAX_OPTION_STRING_SIZE;
  od->cap = SANE_CAP_SOFT_DETECT | SANE_CAP_SOFT_SELECT;
  od->constraint_type = SANE_CONSTRAINT_STRING_LIST;
  od->constraint.string_list = halftone_list;
  lexmark_device->val[LXK_OPT_HALFTONE_MODE].s = malloc(od->size);
  if (!lexmark_device->val[LXK_OPT_HALFTONE_MODE].s)
    return SANE_STATUS_NO_MEM;     
  strcpy(lexmark_device->val[LXK_OPT_HALFTONE_MODE].s,"Error Diffusion");

  /* type of scan */
  od = &(lexmark_device->opt[LXK_OPT_SCAN_TYPE]);
  od->name = "Scan Content";
  od->title = "Scan Content";
  od->desc = "Scan Content";
  od->type = SANE_TYPE_STRING;
  od->unit = SANE_UNIT_NONE;
  od->size = LXK_MAX_OPTION_STRING_SIZE;
  od->cap = SANE_CAP_SOFT_DETECT | SANE_CAP_SOFT_SELECT;
  od->constraint_type = SANE_CONSTRAINT_STRING_LIST;
  od->constraint.string_list = scan_type_list;
  lexmark_device->val[LXK_OPT_SCAN_TYPE].s = malloc(od->size);
  if (!lexmark_device->val[LXK_OPT_SCAN_TYPE].s)
    return SANE_STATUS_NO_MEM;
  strcpy(lexmark_device->val[LXK_OPT_SCAN_TYPE].s,"Photo");

  /* preview mode */
  od = &(lexmark_device->opt[LXK_OPT_PREVIEW]);
  od->name = SANE_NAME_PREVIEW;
  od->title = SANE_TITLE_PREVIEW;
  od->desc = SANE_DESC_PREVIEW;
  od->cap = SANE_CAP_SOFT_DETECT | SANE_CAP_SOFT_SELECT;
  od->type = SANE_TYPE_BOOL;
  lexmark_device->val[LXK_OPT_PREVIEW].w = SANE_FALSE;

  /* dpi group */
  od = &(lexmark_device->opt[LXK_OPT_SCANAREA_GROUP]);
  od->name = "Scan Area";
  od->title = "Scan Area";
  od->desc = "Scan Area";
  od->type = SANE_TYPE_GROUP;

  /* Top-left X */
  od = &(lexmark_device->opt[LXK_OPT_TL_X]);
  od->name = SANE_NAME_SCAN_TL_X;
  od->title = SANE_TITLE_SCAN_TL_X;
  od->desc = SANE_DESC_SCAN_TL_X;
  od->type = SANE_TYPE_INT;
  od->unit = SANE_UNIT_MM;
  od->size = sizeof (SANE_Word);
  od->cap = SANE_CAP_SOFT_DETECT | SANE_CAP_SOFT_SELECT;
  od->constraint_type = SANE_CONSTRAINT_RANGE;
  od->constraint.range = &scan_width_range;
  lexmark_device->val[LXK_OPT_TL_X].w = 0;

  /* Top-left Y */
  od = &(lexmark_device->opt[LXK_OPT_TL_Y]);
  od->name = SANE_NAME_SCAN_TL_Y;
  od->title = SANE_TITLE_SCAN_TL_Y;
  od->desc = SANE_DESC_SCAN_TL_Y;
  od->type = SANE_TYPE_INT;
  od->unit = SANE_UNIT_MM; 
  od->size = sizeof (SANE_Word);
  od->cap = SANE_CAP_SOFT_DETECT | SANE_CAP_SOFT_SELECT;
  od->constraint_type = SANE_CONSTRAINT_RANGE;
  od->constraint.range = &scan_height_range;
  lexmark_device->val[LXK_OPT_TL_Y].w = 0;

  /* Bottom-right X */
  od = &(lexmark_device->opt[LXK_OPT_BR_X]);
  od->name = SANE_NAME_SCAN_BR_X;
  od->title = SANE_TITLE_SCAN_BR_X;
  od->desc = SANE_DESC_SCAN_BR_X;
  od->type = SANE_TYPE_INT;
  od->unit = SANE_UNIT_MM;
  od->size = sizeof (SANE_Word);
  od->cap = SANE_CAP_SOFT_DETECT | SANE_CAP_SOFT_SELECT;
  od->constraint_type = SANE_CONSTRAINT_RANGE;
  od->constraint.range = &scan_width_range;
  lexmark_device->val[LXK_OPT_BR_X].w = scan_width_range.max; //8.5 * 25.4

  /* Bottom-right Y */
  od = &(lexmark_device->opt[LXK_OPT_BR_Y]);
  od->name = SANE_NAME_SCAN_BR_Y;
  od->title = SANE_TITLE_SCAN_BR_Y;
  od->desc = SANE_DESC_SCAN_BR_Y;
  od->type = SANE_TYPE_INT;
  od->unit = SANE_UNIT_MM;
  od->size = sizeof (SANE_Word);
  od->cap = SANE_CAP_SOFT_DETECT | SANE_CAP_SOFT_SELECT;
  od->constraint_type = SANE_CONSTRAINT_RANGE;
  od->constraint.range = &scan_height_range;
  lexmark_device->val[LXK_OPT_BR_Y].w = scan_height_range.max; //11.69 * 25.4

  /*Scan resolution*/
  od = &(lexmark_device->opt[LXK_OPT_DPI]);
  od->name = "Scan resolution";
  od->title = "Scan resolution";
  od->desc = "Scan resolution";
  od->type = SANE_TYPE_INT;
  od->unit = SANE_UNIT_DPI;
  od->size = sizeof (SANE_Word);
  od->cap = SANE_CAP_SOFT_DETECT | SANE_CAP_SOFT_SELECT | SANE_CAP_ADVANCED;
  od->constraint_type = SANE_CONSTRAINT_WORD_LIST;
  od->constraint.word_list = dpi_list;
  lexmark_device->val[LXK_OPT_DPI].w = 150; 
 
  /* dpi button label for high dpi warning*/
  od = &(lexmark_device->opt[LXK_OPT_BUTTON_LABEL1]);
  od->name = "warningtext";
  od->title = "WARNING!!!!                                 \nSelecting higher dpi can cause this \napplication to hang or crash!          \nUse the SAVE mode instead of        \ndisplaying the image in the viewer.  ";
  od->desc = "warningtext";
  od->type = SANE_TYPE_BUTTON;
  od->unit = SANE_UNIT_NONE;
  od->size = sizeof (SANE_Word);
  od->cap = SANE_CAP_SOFT_DETECT | SANE_CAP_SOFT_SELECT | SANE_CAP_ADVANCED | SANE_CAP_INACTIVE;
  //od->cap =  /*SANE_CAP_ADVANCED |*/ SANE_CAP_INACTIVE;
  od->constraint_type = SANE_CONSTRAINT_NONE;
  lexmark_device->val[LXK_OPT_BUTTON_LABEL1].w = 1;

  /* dpi button label for ADF dpi warning*/
  od = &(lexmark_device->opt[LXK_OPT_BUTTON_LABEL2]);
  od->name = "warningtext";
  od->title = "WARNING!!!!\nYou have selected to scan in ADF using an unsupported DPI!\nReverting back to saved supported ADF scan DPI.";
  od->desc = "warningtext";
  od->type = SANE_TYPE_BUTTON;
  od->unit = SANE_UNIT_NONE;
  od->size = sizeof (SANE_Word);
  od->cap = SANE_CAP_SOFT_DETECT | SANE_CAP_SOFT_SELECT | SANE_CAP_ADVANCED | SANE_CAP_INACTIVE;
  //od->cap =  /*SANE_CAP_ADVANCED |*/ SANE_CAP_INACTIVE;
  od->constraint_type = SANE_CONSTRAINT_NONE;
  lexmark_device->val[LXK_OPT_BUTTON_LABEL1].w = 1;

  /* scan source */
  od = &(lexmark_device->opt[LXK_OPT_SCAN_SOURCE]);
  od->name = SANE_NAME_SCAN_SOURCE;
  od->title = SANE_TITLE_SCAN_SOURCE;
  od->desc = SANE_DESC_SCAN_SOURCE;
  od->type = SANE_TYPE_STRING;
  od->unit = SANE_UNIT_NONE;
  od->size = LXK_MAX_OPTION_STRING_SIZE;
  od->cap = SANE_CAP_SOFT_DETECT | SANE_CAP_SOFT_SELECT;
  od->constraint_type = SANE_CONSTRAINT_STRING_LIST;
  od->constraint.string_list = scansource_list;
  lexmark_device->val[LXK_OPT_SCAN_SOURCE].s = malloc(od->size);
  if (!lexmark_device->val[LXK_OPT_SCAN_SOURCE].s)
    return SANE_STATUS_NO_MEM;     
  strcpy(lexmark_device->val[LXK_OPT_SCAN_SOURCE].s,"Flatbed");
  
  return SANE_STATUS_GOOD;
}

/***************************** SANE API ****************************/

SANE_Status
attachLexmark (SANE_String_Const devname)
{
  Lexmark_Device * lexmark_device;
  int fd = -1;
  int pid = 0x00;
  char port[20];
  int index;
  
  DBG (2, "attachLexmark: devname=%s\n", devname);

  for (lexmark_device = first_lexmark_device; lexmark_device; 
       lexmark_device = lexmark_device->next)
  {
    if (strcmp(lexmark_device->sane.name, devname) == 0)
      return SANE_STATUS_GOOD;
  }

  lexmark_device = (Lexmark_Device *) malloc (sizeof(Lexmark_Device));
  if (lexmark_device == NULL)
    return SANE_STATUS_NO_MEM;

  //completely disregard what is passed in devname (always /dev/usb/scanner0). Loop through the rest     
  for(index = 0; index <= 15; index++)
  {
	sprintf(port, "/dev/usb/scanner%d", index);
	fd=open(port, O_WRONLY);
	if(fd >= 0)
	{
		if(!(ioctl(fd, SCANNER_IOCTL_PRODUCT, &pid)))
		{
			close(fd);
			if(pid != 0x98) 
			{
				fprintf(stderr, "\nThe backend installed does not support the attached Lexmark AIO in %s!\nPlease install the appropriate driver for this device.\n", port);
				continue;
			}
			else
				break;
		}
		else
			continue;
	}	
	else continue;
  }
  
  if(index > 15)
	return SANE_STATUS_INVAL;	
  
  lexmark_device->sane.name = strdup (port);
  lexmark_device->sane.vendor = "Lexmark International, Inc."; 
  lexmark_device->sane.model = "X7170"; 
  lexmark_device->sane.type = "All-In-One"; 

  lexmark_device->next = first_lexmark_device;
  first_lexmark_device = lexmark_device;

  num_lexmark_device++;
  
  return SANE_STATUS_GOOD;
}

SANE_Status
sane_init (SANE_Int * version_code, SANE_Auth_Callback authorize)
{
  FILE *fp;
  SANE_Char line[PATH_MAX];
  SANE_String lp;
  SANE_Int vendor, product;
  size_t len;
  SANE_Status status;

  DBG_INIT ();

  DBG (2, "sane_init: version_code=%p, authorize=%p\n", version_code,
       authorize);

  if (version_code)
    *version_code = SANE_VERSION_CODE (V_MAJOR, V_MINOR, BUILD);

  status = llpddk_init ();

  if(status != SANE_STATUS_GOOD)
    return status;
  
  fp = sanei_config_open (LEXMARK_CONFIG_FILE);
  if (!fp)
    {
      llpddk_destroy ();	    
      return SANE_STATUS_ACCESS_DENIED;
    }
  
  while (sanei_config_read (line, PATH_MAX, fp))
    {
      /* ignore comments */
      if (line[0] == '#')
        continue;
      len = strlen (line);

      /* delete newline characters at end */
      if (line[len - 1] == '\n')
        line[--len] = '\0';

      lp = (SANE_String) sanei_config_skip_whitespace (line);

      /* skip empty lines */
      if (*lp == 0)
        continue;

      if (sscanf(lp, "usb %i %i", &vendor, &product) == 2)
        ;
      else if (strncmp("libusb", lp, 6) == 0)
	;      
      else if ((strncmp ("usb", lp, 3) == 0) && isspace (lp[3]))
	{
          lp += 3;
          lp = (SANE_String) sanei_config_skip_whitespace (lp);
        }
      else
        continue;
	  
      llpddk_attach_matching_devices (lp, attachLexmark);
    }
    
  fclose (fp);
  
  initialized = SANE_TRUE;
  
  return SANE_STATUS_GOOD;
}

void
sane_exit (void)
{
  Lexmark_Device * lexmark_device, * next_lexmark_device;

  DBG (2, "sane_exit\n");

  if (!initialized)
    return;
  
  for (lexmark_device = first_lexmark_device; lexmark_device; 
       lexmark_device = next_lexmark_device)
    {
      next_lexmark_device = lexmark_device->next;
      free (lexmark_device);
    }  

  if (sane_device_list)
    free (sane_device_list);

  llpddk_destroy ();

  initialized = SANE_FALSE;
  
  return;
}

SANE_Status
sane_get_devices (const SANE_Device *** device_list, SANE_Bool local_only)
{
  Lexmark_Device * lexmark_device;
  SANE_Int index;
  
  DBG (2, "sane_get_devices: device_list=%p, local_only=%d\n",
       device_list, local_only);

  if (!initialized)
    return SANE_STATUS_INVAL;
  
  if (sane_device_list)
    free (sane_device_list);

  sane_device_list = (SANE_Device **) calloc (num_lexmark_device+1,
		                              sizeof(sane_device_list[0]));
  if (!sane_device_list)
    return SANE_STATUS_NO_MEM;

  index = 0;
  for (lexmark_device = first_lexmark_device; lexmark_device; 
       lexmark_device = lexmark_device->next)
    {
      sane_device_list[index] = &(lexmark_device->sane);
      index++;
    }
  sane_device_list[index] = 0;

  *device_list = sane_device_list;
  
  return SANE_STATUS_GOOD;
}

SANE_Status
sane_open (SANE_String_Const devicename, SANE_Handle * handle)
{
  Lexmark_Device * lexmark_device;
  SANE_Status status;

  DBG (2, "sane_open: devicename=\"%s\", handle=%p\n", devicename, handle);

  if (!initialized)
    return SANE_STATUS_INVAL;
  
  if (!handle)
    return SANE_STATUS_INVAL;

  for (lexmark_device = first_lexmark_device; lexmark_device;
       lexmark_device = lexmark_device->next)
    {
      if (strcmp (devicename, lexmark_device->sane.name) == 0)
        break;
    }
  
  *handle = lexmark_device;

  if (!lexmark_device)
    return SANE_STATUS_INVAL;
  
  status = init_options (lexmark_device);
  if (status != SANE_STATUS_GOOD)
    return status;

  status = llpddk_open_device (lexmark_device->sane.name);
  if (status != SANE_STATUS_GOOD)
    return status;
 
  return SANE_STATUS_GOOD;
}

void
sane_close (SANE_Handle handle)
{
  Lexmark_Device * lexmark_device;

  DBG (2, "sane_close: handle=%p\n", handle);

  if (!initialized)
    return;
  
  for (lexmark_device = first_lexmark_device; lexmark_device;
       lexmark_device = lexmark_device->next)
    {
      if (lexmark_device == handle)
        break;
    }
  
  if (!lexmark_device)
    return;
 
  llpddk_close_device ();

  return;
}

const SANE_Option_Descriptor *
sane_get_option_descriptor (SANE_Handle handle, SANE_Int option)
{
  Lexmark_Device * lexmark_device;

  DBG (2, "sane_get_option_descriptor: handle=%p, option = %d\n",  
       handle, option);

  if (!initialized)
    return NULL;
  
  for (lexmark_device = first_lexmark_device; lexmark_device;
       lexmark_device = lexmark_device->next)
    {
      if (lexmark_device == handle)
        break;
    }
  
  if (!lexmark_device)
    return NULL;

  return &(lexmark_device->opt[option]);
}

SANE_Status
sane_control_option (SANE_Handle handle, SANE_Int option, SANE_Action action,
		     void * value, SANE_Int * info)
{
  Lexmark_Device * lexmark_device;
  SANE_Status status;

  DBG (2, "sane_control_option: handle=%p, opt=%d, act=%d, val=%p, info=%p\n",
       handle, option, action, value, info);

  if (!initialized)
    return SANE_STATUS_INVAL;
  
  for (lexmark_device = first_lexmark_device; lexmark_device;
       lexmark_device = lexmark_device->next)
    {
      if (lexmark_device == handle)
        break;
    }
  
  if (!lexmark_device)
    return SANE_STATUS_INVAL;
   
  if (value == NULL && lexmark_device->opt[option].type != SANE_TYPE_BUTTON)
    return SANE_STATUS_INVAL;

  if (info != NULL)
    *info = 0;
  
  if (option < 0 || option >= LXK_NUM_OPTIONS)
    return SANE_STATUS_INVAL;

  //if (lexmark_device->opt[option].type == SANE_TYPE_GROUP)
  //  return SANE_STATUS_INVAL;
    
  switch (action)
  {
    case SANE_ACTION_SET_AUTO:

      if (!SANE_OPTION_IS_SETTABLE (lexmark_device->opt[option].cap))
	      return SANE_STATUS_INVAL;
      if (!(lexmark_device->opt[option].cap & SANE_CAP_AUTOMATIC))
	      return SANE_STATUS_INVAL;
      break;

    case SANE_ACTION_SET_VALUE:

      if (!SANE_OPTION_IS_SETTABLE (lexmark_device->opt[option].cap))
	      return SANE_STATUS_INVAL;

      switch(option) 
      {
	        case LXK_OPT_NUM_OPTS:
          	case LXK_OPT_DPI:
	        {
				lexmark_device->val[option].w = *(SANE_Int *) value;
	            if(!strcmp(lexmark_device->val[LXK_OPT_SCAN_SOURCE].s, scansource_list[1])) //if scan source is ADF
				{
					if(*(SANE_Int *) value > 300)
					{
						lexmark_device->opt[LXK_OPT_BUTTON_LABEL1].cap |= SANE_CAP_INACTIVE;
			    		lexmark_device->opt[LXK_OPT_BUTTON_LABEL2].cap &= 0xdf;
						lexmark_device->val[option].w = oldDPI;
					}
					else
					{
						oldDPI = lexmark_device->val[option].w; 
			    		lexmark_device->opt[LXK_OPT_BUTTON_LABEL2].cap |= SANE_CAP_INACTIVE; 
					}						
				}
				else
				{
					if(*(SANE_Int *) value > 600)
					{
			    		lexmark_device->opt[LXK_OPT_BUTTON_LABEL1].cap &= 0xdf;
						lexmark_device->opt[LXK_OPT_BUTTON_LABEL2].cap |= SANE_CAP_INACTIVE;
					}
					else
			    		lexmark_device->opt[LXK_OPT_BUTTON_LABEL1].cap |= SANE_CAP_INACTIVE; 
				}
			
				if (info)
			    	*info |= SANE_INFO_RELOAD_OPTIONS;
		    	sane_get_parameters(handle, 0);
		    	
				if (info)
			    	*info |= SANE_INFO_RELOAD_OPTIONS;
				
		    					
				return SANE_STATUS_GOOD;
		       	break;
			}
	       	case LXK_OPT_TL_X:
	        case LXK_OPT_TL_Y:
	        case LXK_OPT_BR_X:
	        case LXK_OPT_BR_Y:
	            lexmark_device->val[option].w = *(SANE_Int *) value;
              	break;	  
		  	case LXK_OPT_BUTTON_LABEL1:  
			    lexmark_device->opt[LXK_OPT_BUTTON_LABEL1].cap |= SANE_CAP_INACTIVE; 
			    if (info)
				    *info |= SANE_INFO_RELOAD_OPTIONS;
			    sane_get_parameters(handle, 0);
		    	if (info)
			    	*info |= SANE_INFO_RELOAD_PARAMS;
		  	    return SANE_STATUS_GOOD;
			case LXK_OPT_BUTTON_LABEL2:  
			    lexmark_device->opt[LXK_OPT_BUTTON_LABEL2].cap |= SANE_CAP_INACTIVE; 
			    if (info)
				    *info |= SANE_INFO_RELOAD_OPTIONS;
			    sane_get_parameters(handle, 0);
		    	if (info)
			    	*info |= SANE_INFO_RELOAD_PARAMS;
		  	    return SANE_STATUS_GOOD;
			case LXK_OPT_PREVIEW:
				lexmark_device->val[option].w = *(SANE_Int *) value;

				if (*(SANE_Word *) value) 
				{
					dpi = lexmark_device->val[LXK_OPT_DPI].w;
					lexmark_device->val[LXK_OPT_DPI].w = 75;
					lexmark_device->val[LXK_OPT_TL_X].w = 0;
					lexmark_device->val[LXK_OPT_TL_Y].w = 0;					
					lexmark_device->val[LXK_OPT_BR_X].w = 216;  	//8.5 inches - letter's width
					lexmark_device->val[LXK_OPT_BR_Y].w = 297;		//11.69 inches -A4's height
				}
				else
				{
					lexmark_device->val[LXK_OPT_DPI].w = dpi;
					lexmark_device->val[LXK_OPT_TL_X].w = 0;
					lexmark_device->val[LXK_OPT_TL_Y].w = 0;					
					lexmark_device->val[LXK_OPT_BR_X].w = 216;	
					lexmark_device->val[LXK_OPT_BR_Y].w = 297;	
				}

				if(!paperJamOnADF)
					getADFStatus();

				if(paperOnADF)
				{
					lexmark_device->val[LXK_OPT_BR_Y].w = 356;	//14 inches - Legal's height
					startADF = SANE_FALSE;
				}

	    		if (info)
	      			*info |= SANE_INFO_RELOAD_OPTIONS;	
				sane_get_parameters (handle, 0);
	  			if (info)
	    			*info |= SANE_INFO_RELOAD_PARAMS;
             	break;	  
          case LXK_OPT_SCAN_TYPE:
	          	strcpy(lexmark_device->val[option].s, value);
            	if(!(strcmp(lexmark_device->val[option].s, scan_type_list[1])))                   
              		lexmark_device->val[LXK_OPT_DPI].w = 300;
             	else
     				lexmark_device->val[LXK_OPT_DPI].w = 150;

         		lexmark_device->opt[LXK_OPT_BUTTON_LABEL1].cap |= SANE_CAP_INACTIVE; 

            	if (info)
			        *info |= SANE_INFO_RELOAD_OPTIONS;
  		      	sane_get_parameters(handle, 0);
	  	      	if (info)
		  	      	*info |= SANE_INFO_RELOAD_OPTIONS;
		        break;
          case LXK_OPT_CHANNELS:
          case LXK_OPT_HALFTONE_MODE:
			   strcpy(lexmark_device->val[option].s, value);
		  	   break;
		  case LXK_OPT_SCAN_SOURCE:  
			    if(!paperJamOnADF)
					getADFStatus();
		  		if(paperOnADF)
				{
	            	strcpy(lexmark_device->val[option].s, scansource_list[1]); //force it to ADF
					lexmark_device->val[LXK_OPT_TL_X].w = 0;
					lexmark_device->val[LXK_OPT_TL_Y].w = 0;					
					lexmark_device->val[LXK_OPT_BR_X].w = 216;	//8.5 in
					lexmark_device->val[LXK_OPT_BR_Y].w = 356;	//14 in	
					lexmark_device->opt[LXK_OPT_BUTTON_LABEL1].cap |= SANE_CAP_INACTIVE; //disable any previous high dpi warning
					if(lexmark_device->val[LXK_OPT_DPI].w  > 300)
					{
						lexmark_device->opt[LXK_OPT_BUTTON_LABEL1].cap |= SANE_CAP_INACTIVE;
			    		lexmark_device->opt[LXK_OPT_BUTTON_LABEL2].cap &= 0xdf;
						lexmark_device->val[LXK_OPT_DPI].w = oldDPI;
					}
					else
			    		lexmark_device->opt[LXK_OPT_BUTTON_LABEL2].cap |= SANE_CAP_INACTIVE;					
				}
				else
				{
					lexmark_device->val[LXK_OPT_TL_X].w = 0;
					lexmark_device->val[LXK_OPT_TL_Y].w = 0;					
					lexmark_device->val[LXK_OPT_BR_X].w = 216;	//8.5 in
					lexmark_device->val[LXK_OPT_BR_Y].w = 297;	//11.67 in	
					strcpy(lexmark_device->val[option].s, scansource_list[0]); //force it to flatbed
					lexmark_device->opt[LXK_OPT_BUTTON_LABEL2].cap |= SANE_CAP_INACTIVE; //disable any previous ADF unsupported DPI warning
					if(startADF)
						startADF = SANE_FALSE;
				}
				
				if (info)			     
					*info |= SANE_INFO_RELOAD_OPTIONS;
  		      	sane_get_parameters(handle, 0);
	  	      	if (info)
		  	      *info |= SANE_INFO_RELOAD_PARAMS;
	            break;	  
      } 

      if (info != NULL)
          *info |= SANE_INFO_RELOAD_PARAMS;

      break;

    case SANE_ACTION_GET_VALUE:

      switch (option)
      {
		    case LXK_OPT_NUM_OPTS: 
          	case LXK_OPT_DPI:
	        case LXK_OPT_TL_X:
	        case LXK_OPT_TL_Y:
	        case LXK_OPT_BR_X:
	        case LXK_OPT_BR_Y:
			case LXK_OPT_PREVIEW:
              		*(SANE_Int *) value = lexmark_device->val[option].w;
              		break;
 	        case LXK_OPT_CHANNELS:
          	case LXK_OPT_HALFTONE_MODE:
			case LXK_OPT_SCAN_SOURCE:
          	case LXK_OPT_SCAN_TYPE: 
				strcpy(value, lexmark_device->val[option].s);
			 	break;				
		  	case LXK_OPT_BUTTON_LABEL1:
			case LXK_OPT_BUTTON_LABEL2:
				break;
	      	default:
      	      	return SANE_STATUS_INVAL;
	    }

	    break;
      
    	   default:
		return SANE_STATUS_INVAL;
    }
  
  return SANE_STATUS_GOOD;
}


SANE_Status
sane_get_parameters (SANE_Handle handle, SANE_Parameters * params)
{
	Lexmark_Device * lexmark_device;
  	SANE_Parameters * device_params;
  	SANE_Int res, width_px, height_px;
  	SANE_Int channels, bitsperchannel;
  	double ulx, uly, brx, bry;
  	unsigned char adfStatus;
	
  	DBG (2, "sane_get_parameters: handle=%p, params=%p\n", handle, params);

  	if (!initialized)
    	return SANE_STATUS_INVAL;

  for (lexmark_device = first_lexmark_device; lexmark_device;
       lexmark_device = lexmark_device->next)
  {
      if (lexmark_device == handle)
        break;
  }
  
  if (!lexmark_device)
    return SANE_STATUS_INVAL;
  
  res = lexmark_device->val[LXK_OPT_DPI].w;   

  //top left x
  ulx = (double) (lexmark_device->val[LXK_OPT_TL_X].w / MM_PER_INCH); 
  ulx = (SANE_Int) (ulx*res); //ulx now in px;

  //top left y
  uly = (double) (lexmark_device->val[LXK_OPT_TL_Y].w / MM_PER_INCH);
  uly = (SANE_Int) (uly*res); //uly now in pix;

  //bottom right x
  brx = (double) (lexmark_device->val[LXK_OPT_BR_X].w / MM_PER_INCH); 
  brx = (SANE_Int) (brx*res); //ulx now in px;
 
  //bottom right y
  bry = (double) (lexmark_device->val[LXK_OPT_BR_Y].w / MM_PER_INCH);
  bry = (SANE_Int) (bry*res); //uly now in pix;

  /*if(res > 1200)
  {
  	ulx = (SANE_Int) (ulx/16); ulx = (SANE_Int) (ulx*16); 
  	uly = (SANE_Int) (uly/16); uly = (SANE_Int) (uly*16);     
  	brx = (SANE_Int) (brx/16); brx = (SANE_Int) (brx*16);  
  	bry = (SANE_Int) (bry/16); bry = (SANE_Int) (bry*16);  
  }*/

  if(res > 600)
  {
  	ulx = (SANE_Int) (ulx/32); ulx = (SANE_Int) (ulx*32); 
  	uly = (SANE_Int) (uly/32); uly = (SANE_Int) (uly*32);     
  	brx = (SANE_Int) (brx/32); brx = (SANE_Int) (brx*32);  
  	bry = (SANE_Int) (bry/32); bry = (SANE_Int) (bry*32);  
  }

  width_px = (SANE_Int) (brx - ulx);
  height_px = (SANE_Int) (bry - uly);

  channels =  3;
  bitsperchannel = 8;
  
  if (strcmp(lexmark_device->val[LXK_OPT_CHANNELS].s,"Gray") == 0)
	channels = 1;
 
  data_size = width_px*height_px*channels;

  device_params = &(lexmark_device->params);  
  device_params->format = SANE_FRAME_RGB;
  if (channels == 1)
    device_params->format = SANE_FRAME_GRAY;
  	device_params->lines = height_px;	
  	device_params->depth = bitsperchannel;  
  	device_params->pixels_per_line =  width_px; 
  	device_params->bytes_per_line = (SANE_Int) (channels * 
  	device_params->pixels_per_line);//since bith_depth for both gray and color is both 8
  	device_params->last_frame = SANE_TRUE; 
  
  if (params)
  {  
     params->format = device_params->format;
     params->last_frame = device_params->last_frame;
     params->lines = device_params->lines;
     params->depth = device_params->depth;
     params->pixels_per_line = device_params->pixels_per_line;
     params->bytes_per_line = device_params->bytes_per_line;
  }
	
	return SANE_STATUS_GOOD;
}

SANE_Status
sane_start (SANE_Handle handle)
{
  Lexmark_Device * lexmark_device;  
  ssize_t bytes_read;
  unsigned char data[8];
	
  DBG (2, "sane_start: handle=%p\n", handle);
  
  if (!initialized)
    return SANE_STATUS_INVAL;

  for (lexmark_device = first_lexmark_device; lexmark_device;
       lexmark_device = lexmark_device->next)
    {
      if (lexmark_device == handle)
        break;
    }
  
  if (!lexmark_device)
    return SANE_STATUS_INVAL;
  sane_get_parameters (handle, 0);
  getADFStatus();
  if(paperJamOnADF) //paper jam on ADF, abort scan
  {
	  paperJamOnADF = SANE_FALSE;
  	  return SANE_STATUS_JAMMED;
  }
  
  if(!strcmp(lexmark_device->val[LXK_OPT_SCAN_SOURCE].s, scansource_list[1]) && !paperOnADF)
  {
	  //invoking ADF scan but no paper in ADF
	  startADF = SANE_FALSE;
	  //set data_ctr equal to data_size so as not to satisfy condition in sane_cancel to read remaining data
	  data_ctr = data_size;
	  return SANE_STATUS_NO_DOCS;
  }	  
  else if(!strcmp(lexmark_device->val[LXK_OPT_SCAN_SOURCE].s, scansource_list[0]) && paperOnADF)
  {
	  //force ADF scanning if there is paper and scan source selected is flatbed
	  strcpy(lexmark_device->val[LXK_OPT_SCAN_SOURCE].s, scansource_list[1]);
	  forcedADF = SANE_TRUE;
	  lexmark_device->val[LXK_OPT_DPI].w = oldDPI;
	  lexmark_device->opt[LXK_OPT_BUTTON_LABEL1].cap |= SANE_CAP_INACTIVE;
	  lexmark_device->opt[LXK_OPT_BUTTON_LABEL2].cap |= SANE_CAP_INACTIVE;
  } 
  
  if(paperOnADF  && !startADF)  //check for ADF scanning
  {
	  startADF = SANE_TRUE;
	  lexmark_device->val[LXK_OPT_TL_X].w = 0;
	  lexmark_device->val[LXK_OPT_TL_Y].w = 0;					
	  lexmark_device->val[LXK_OPT_BR_X].w = 216;  //8.5
	  lexmark_device->val[LXK_OPT_BR_Y].w = 356;  //14	
  }
  else if(startADF && !paperOnADF)
  {
	  startADF = SANE_FALSE;
	  return SANE_STATUS_NO_DOCS;  
  }
  
  if ((lexmark_device->params.lines == 0) ||
      (lexmark_device->params.pixels_per_line == 0) ||
      (lexmark_device->params.bytes_per_line == 0))
      return SANE_STATUS_INVAL;
 	device_cancelled = SANE_FALSE;
	gui_cancelled = SANE_FALSE;
	data_ctr = 0;

  	//need this cancel_ctr to determine how many times sane_cancel is called 
  	//since it is called more than once in xscanimage's preview
  	cancel_ctr = 0; 

  if(llpddk_start_scan(lexmark_device->val) == SANE_STATUS_GOOD)
 	  return SANE_STATUS_GOOD;
   else
  {
  	  device_cancelled = SANE_TRUE;
	  getADFStatus();
	  if(paperJamOnADF)
		  return SANE_STATUS_JAMMED;
	  return SANE_STATUS_INVAL;
  }	  
}


SANE_Status
sane_read (SANE_Handle handle, SANE_Byte * data,
	   SANE_Int max_length, SANE_Int * length)
{
  Lexmark_Device * lexmark_device;
  ssize_t bytes_read;

  DBG (2, "sane_read: handle=%p, data=%p, max_length = %d, length=%p\n",
       handle, data, max_length, length);

  if (!initialized)
    return SANE_STATUS_INVAL;
  if(gui_cancelled == SANE_TRUE)
    return SANE_STATUS_CANCELLED;
  
  for (lexmark_device = first_lexmark_device; lexmark_device;
       lexmark_device = lexmark_device->next)
    {
      if (lexmark_device == handle)
        break;
    }
  
  if (!lexmark_device)
    return SANE_STATUS_INVAL;
  
  if (!length)
    return SANE_STATUS_INVAL;

  *length = 0;
   
  if (eof)
   return SANE_STATUS_EOF;

  if (!data)
    return SANE_STATUS_INVAL;
  
  bytes_read = (size_t) llpddk_read_scan_data (data, max_length, &eof);
  if (bytes_read < 0)
  {
    if(bytes_read == IO_READ_ERROR)
	{
		//COMMENT TO BE REMOVED IN FINAL RELEASE_LOCK
		//this is work-around for PTR 211331 - unexpected output if usb cable is detached
		//this only happens at first scan after cable is detached, succeeding scan is ok
		//to solve this, we force quit the application to start with a clean scan 
		exit(0);
		//return SANE_STATUS_IO_ERROR;
	}
	else if(bytes_read == ADF_PAPER_JAM)
	{
		paperJamOnADF = SANE_TRUE;
		if(startADF)
		{
			if(forcedADF)
			{
				strcpy(lexmark_device->val[LXK_OPT_SCAN_SOURCE].s, scansource_list[0]);	//set back to flatbed scan
				forcedADF = SANE_FALSE;
	 			lexmark_device->val[LXK_OPT_TL_X].w = 0;
				lexmark_device->val[LXK_OPT_TL_Y].w = 0;					
				lexmark_device->val[LXK_OPT_BR_X].w = 216;  	//8.5 inches - letter's width
				lexmark_device->val[LXK_OPT_BR_Y].w = 297;		//11.69 inches -A4's height
			}
			startADF = SANE_FALSE;
		}
		return SANE_STATUS_JAMMED;
	}
  }
  else if(bytes_read == 0)
  {
	getADFStatus();
    if(startADF && paperOnADF)
		sane_cancel(handle);
	else if(startADF && !paperOnADF)
	{
		startADF = SANE_FALSE;
		if(forcedADF)
		{
			strcpy(lexmark_device->val[LXK_OPT_SCAN_SOURCE].s, scansource_list[0]);	//set back to flatbed scan
			forcedADF = SANE_FALSE;
			lexmark_device->val[LXK_OPT_TL_X].w = 0;
			lexmark_device->val[LXK_OPT_TL_Y].w = 0;					
			lexmark_device->val[LXK_OPT_BR_X].w = 216;  	//8.5 inches - letter's width
			lexmark_device->val[LXK_OPT_BR_Y].w = 297;		//11.69 inches -A4's height
		}
    }
	
	
	return SANE_STATUS_EOF;
  }
  else if(bytes_read == 8 && data[0] == 0x1b && data[1] == 0x53 && data[2] == 0x53 && data[3] == 0x09)
  {
    	device_cancelled = SANE_TRUE;
		if(startADF)
		{
			startADF = SANE_FALSE;
    		if(forcedADF)
			{
				strcpy(lexmark_device->val[LXK_OPT_SCAN_SOURCE].s, scansource_list[0]);	//set back to flatbed scan
				forcedADF = SANE_FALSE;
				lexmark_device->val[LXK_OPT_TL_X].w = 0;
				lexmark_device->val[LXK_OPT_TL_Y].w = 0;					
				lexmark_device->val[LXK_OPT_BR_X].w = 216;  	//8.5 inches - letter's width
				lexmark_device->val[LXK_OPT_BR_Y].w = 297;		//11.69 inches -A4's height
			}
		}
		return SANE_STATUS_CANCELLED;
  }
  else
  {
    *length = bytes_read;
    data_ctr += bytes_read;
	if((signed int) data_ctr < 0)
	{
		fprintf(stderr, "Scanner data exceeded maximum, aborting..\n");
		llpddk_report_error(LXK_MEMORY_OUT);
		llpddk_abort_scan();
	 	do
	 		bytes_read = (size_t) llpddk_read_scan_data (data, sizeof(data), &eof);
	 	while(bytes_read);
		llpddk_end_scan();			
    }
  }

  return SANE_STATUS_GOOD;  
}

void
sane_cancel (SANE_Handle handle)
{
  Lexmark_Device * lexmark_device;
  ssize_t bytes_read;
  char port[20];
  int index;
  int fd = -1;
  int pid = 0x00;
	
  DBG (2, "sane_cancel: handle = %p\n", handle);
	
  eof = 0;

  if (!initialized)
    return;
  
  if(++cancel_ctr > 1)  //sane_cancel called more than once, return
	 return;

  for (lexmark_device = first_lexmark_device; lexmark_device;
       lexmark_device = lexmark_device->next)
  {
	if (lexmark_device == handle)
        	break;
  }
  
   if (!lexmark_device)
   {
	   DBG (2, "sane_cancel::lexmark_device==NULL\n"); 
       return;
   }

   if(paperJamOnADF)
	   return;
	
   //check if device is still connected, return if not
   if(system("grep 7100 /proc/bus/usb/devices >/dev/null 2>&1"))
   {
	   usbDetached = SANE_TRUE;
	   //llpddk_end_scan();  //fails if logged as user
	   //llpddk_report_error(LXK_COMMUNICATION_ERROR);
	   return;	   
   }

    if((data_ctr < data_size) && !device_cancelled) 
    {
		if(forcedADF)
		{
			strcpy(lexmark_device->val[LXK_OPT_SCAN_SOURCE].s, scansource_list[0]);	//set back to flatbed scan
			forcedADF = SANE_FALSE;
			lexmark_device->val[LXK_OPT_TL_X].w = 0;
			lexmark_device->val[LXK_OPT_TL_Y].w = 0;					
			lexmark_device->val[LXK_OPT_BR_X].w = 216;  	//8.5 inches - letter's width
			lexmark_device->val[LXK_OPT_BR_Y].w = 297;		//11.69 inches -A4's height
		}
		
		if(startADF)
			startADF = SANE_FALSE;	
		
		unsigned char data[32768];
   	 	llpddk_abort_scan();
	 	do
	 		bytes_read = (size_t) llpddk_read_scan_data (data, sizeof(data), &eof);
	 	while(bytes_read);
  	 	gui_cancelled = SANE_TRUE;
    }
   	llpddk_end_scan();
    	device_cancelled = SANE_FALSE;
   	eof = 0;
	
    return;
}

SANE_Status
sane_set_io_mode (SANE_Handle handle, SANE_Bool non_blocking)
{
  Lexmark_Device * lexmark_device;

  DBG (2, "sane_set_io_mode: handle = %p, non_blocking = %d\n", handle,
       non_blocking);

  if (!initialized)
    return SANE_STATUS_INVAL;
  
  for (lexmark_device = first_lexmark_device; lexmark_device;
       lexmark_device = lexmark_device->next)
    {
      if (lexmark_device == handle)
        break;
    }
  
  if (!lexmark_device)
    return SANE_STATUS_INVAL;
 
  if (non_blocking)                       
    return SANE_STATUS_UNSUPPORTED;

  return SANE_STATUS_GOOD;
}

SANE_Status
sane_get_select_fd (SANE_Handle handle, SANE_Int * fd)
{
  Lexmark_Device * lexmark_device;

  DBG (2, "sane_get_select_fd: handle = %p, fd %s 0\n", handle,
       fd ? "!=" : "=");

  if (!initialized)
    return SANE_STATUS_INVAL;
  
  for (lexmark_device = first_lexmark_device; lexmark_device;
       lexmark_device = lexmark_device->next)
    {
      if (lexmark_device == handle)
        break;
    }
  
  if (!lexmark_device)
    return SANE_STATUS_INVAL;
 
  return SANE_STATUS_UNSUPPORTED;
}

/***************************** END OF SANE API ****************************/

