/**************************************************************************
   llpddk.cpp - SANE library for Lexmark scanners.
   Copyright (C) 2003-2005 Lexmark International, Inc.
 
   This file is part of the SANE package.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.

   As a special exception, the authors of SANE give permission for
   additional uses of the libraries contained in this release of SANE.

   The exception is that, if you link a SANE library with other files
   to produce an executable, this does not by itself cause the
   resulting executable to be covered by the GNU General Public
   License.  Your use of that executable is in no way restricted on
   account of linking the SANE library code into it.

   This exception does not, however, invalidate any other reasons why
   the executable file might be covered by the GNU General Public
   License.

   If you submit changes to SANE to the maintainers to be included in
   a subsequent release, you agree by submitting the changes that
   those changes may be distributed with this exception intact.

   If you write modifications of your own for SANE, it is your choice
   whether to permit this exception to apply to your modifications.
   If you do not wish that, delete this exception notice.
   **************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <lexmark-H/scannerdevice.h>
#include "../include/sane/sane.h"
#include "../backend/lexmark.h"

#include "saneportmonitor.h"
#include "scanerrorcommunicator.h"

#define MM_PER_INCH 25.4

ScannerDevice *sd;
SanePortMonitor *sanepm;
ScanErrorCommunicator * ei;

extern "C" SANE_Status llpddk_init()
{
	sanepm = new SanePortMonitor;
	if (sanepm == NULL)
		return SANE_STATUS_INVAL;

  	ei = new ScanErrorCommunicator;
	
	if (ei == NULL)
	{
    	delete sanepm;
    	return SANE_STATUS_INVAL;        
  	}
  
  	sd = new ScannerDevice(sanepm, ei, "X7170");

  	if (sd == NULL)
	{
		delete sanepm;
    	delete ei;
		return SANE_STATUS_INVAL;
	}
	
	return SANE_STATUS_GOOD;	
}

extern "C" void llpddk_attach_matching_devices(
	const char * name,
	SANE_Status (* attach)(const char * dev))
{
	sanepm->PM_AttachDevice(name, attach);

	return;
}

extern "C" void llpddk_destroy(
	void )
{

	if (sd)
		delete sd;
	
	if (ei)
	  delete ei;

	if (sanepm)
		delete sanepm;

	return;
}

extern "C" SANE_Status llpddk_open_device(
	SANE_String_Const devname)
{
	PortMonitor::PM_Error pmerr;
	
	pmerr = sanepm->PM_SetDeviceName(devname);
	if (pmerr != PortMonitor::PM_ERR_SUCCESS)
		return SANE_STATUS_INVAL;

	pmerr = sanepm->PM_Open();
	if (pmerr != PortMonitor::PM_ERR_SUCCESS)
		return SANE_STATUS_INVAL;

	return SANE_STATUS_GOOD;
}

extern "C" void llpddk_close_device(
	void)
{
	sanepm->PM_Close();
	
	return;
}

extern "C" SANE_Status llpddk_start_scan(
	Option_Value opt[])
{
	ScannerDevice::SD_ErrCode sderr;
	ScannerDevice::SD_ScanProp sdprop;

 	// bits per channel
	sdprop.bitsperchannel = ScannerDevice::SD_EIGHT_BPC;
	
	// channels
	sdprop.channels = ScannerDevice::SD_RGB_CHANNEL;

	if (strcmp(opt[LXK_OPT_CHANNELS].s,"Gray") == 0)
		sdprop.channels = ScannerDevice::SD_MONO_CHANNEL;

	// halftone
	sdprop.halftonemode = ScannerDevice::SD_ERROR_DIFFUSION;
	if (strcmp(opt[LXK_OPT_HALFTONE_MODE].s,"Threshold") == 0) 
		sdprop.halftonemode = ScannerDevice::SD_THRESHOLD;

  // content type / type of scan
  if (strcmp(opt[LXK_OPT_SCAN_TYPE].s,"Photo") == 0)
		sdprop.contenttype = ScannerDevice::SD_PHOTO;
  else if (strcmp(opt[LXK_OPT_SCAN_TYPE].s,"Text") == 0)
    sdprop.contenttype = ScannerDevice::SD_TEXT_DOCUMENT;
  else
    sdprop.contenttype = ScannerDevice::SD_MIXED;
    
	// compression
	sdprop.compressionmode = ScannerDevice::SD_UNCOMPRESSED;

	// dpi
	unsigned int dpi;
	switch (opt[LXK_OPT_DPI].w)
	{
		case 75: sdprop.horizontaldpi = sdprop.verticaldpi = ScannerDevice::SD_75_DPI;
			 dpi = 75;
			 break;
		case 100: sdprop.horizontaldpi = sdprop.verticaldpi = ScannerDevice::SD_100_DPI;
			  dpi = 100;
			  break;
		case 150: sdprop.horizontaldpi = sdprop.verticaldpi = ScannerDevice::SD_150_DPI;
			  dpi = 150;
			  break;
		case 300: sdprop.horizontaldpi = sdprop.verticaldpi = ScannerDevice::SD_300_DPI;
			  dpi = 300;
			  break;
		case 600: sdprop.horizontaldpi = sdprop.verticaldpi = ScannerDevice::SD_600_DPI;
			  dpi = 600;
			  break;
		case 1200: sdprop.horizontaldpi = sdprop.verticaldpi = ScannerDevice::SD_1200_DPI;
			  dpi = 1200;
			   break;
		case 2400: sdprop.horizontaldpi = sdprop.verticaldpi = ScannerDevice::SD_2400_DPI;
			  dpi = 2400;
			   break;
		case 4800: sdprop.horizontaldpi = sdprop.verticaldpi = ScannerDevice::SD_4800_DPI;
			  dpi = 4800;
			   break;
		case 9600: sdprop.horizontaldpi = sdprop.verticaldpi = ScannerDevice::SD_9600_DPI;
			  dpi = 9600;
			   break;
		case 19200: sdprop.horizontaldpi = sdprop.verticaldpi = ScannerDevice::SD_19200_DPI;
			  dpi = 19200;
			   break;
		default: sdprop.horizontaldpi = sdprop.verticaldpi = ScannerDevice::SD_150_DPI;
			 break;
	}
	double tl_x, tl_y, br_x, br_y;
	
	//to pixels
	tl_x = (double) (opt[LXK_OPT_TL_X].w / MM_PER_INCH);
	tl_x = (unsigned int) (tl_x*dpi);
	
	tl_y = (double) (opt[LXK_OPT_TL_Y].w / MM_PER_INCH);
	tl_y = (unsigned int) (tl_y*dpi);
	
	br_x = (double) (opt[LXK_OPT_BR_X].w / MM_PER_INCH);
	br_x = (unsigned int) (br_x*dpi);
	
	br_y = (double) (opt[LXK_OPT_BR_Y].w / MM_PER_INCH);
	br_y = (unsigned int) (br_y*dpi);
	
	/*if(opt[LXK_OPT_DPI].w > 1200)
	{
	tl_x = (unsigned int) (tl_x/16); tl_x = (unsigned int) (tl_x*16);
	tl_y = (unsigned int) (tl_y/16); tl_y = (unsigned int) (tl_y*16);
	br_x = (unsigned int) (br_x/16); br_x = (unsigned int) (br_x*16);
	br_y = (unsigned int) (br_y/16); br_y = (unsigned int) (br_y*16);	
	}*/
    
	//let us make the width and height divisible by 32 to be uniform with values in the ldk	
	if(opt[LXK_OPT_DPI].w > 600) 
	{
	tl_x = (unsigned int) (tl_x/32); tl_x = (unsigned int) (tl_x*32);
	tl_y = (unsigned int) (tl_y/32); tl_y = (unsigned int) (tl_y*32);
	br_x = (unsigned int) (br_x/32); br_x = (unsigned int) (br_x*32);
	br_y = (unsigned int) (br_y/32); br_y = (unsigned int) (br_y*32);	
	}
	
	sdprop.horizontaloffset = tl_x;
	sdprop.verticaloffset = tl_y;
	sdprop.width = br_x - tl_x;
	sdprop.height = br_y - tl_y;
	
	sderr = sd->SD_StartScan(sdprop);
	if (sderr != ScannerDevice::SD_OK)
		return SANE_STATUS_INVAL;

	return SANE_STATUS_GOOD;
}

extern "C" long llpddk_read_scan_data(
	SANE_Byte * data,
	SANE_Int size,
	int * eof)
{	
	return sd->SD_ReadScanData(
		(unsigned char *) data,
		(unsigned long) size,
		eof);
}

extern "C" SANE_Status llpddk_abort_scan(
	void)
{
	ScannerDevice::SD_ErrCode sderr;

 
	sderr = sd->SD_CancelScan();
	if (sderr != ScannerDevice::SD_OK)
		return SANE_STATUS_INVAL;

	return SANE_STATUS_GOOD;
}

extern "C" SANE_Status llpddk_end_scan(
	void)
{
	ScannerDevice::SD_ErrCode sderr;

 
	sderr = sd->SD_EndScan();
	if (sderr != ScannerDevice::SD_OK)
		return SANE_STATUS_INVAL;

	return SANE_STATUS_GOOD;
}
 
extern "C" SANE_Status llpddk_get_adf_status(
	unsigned char * adfStat)
{
	ScannerDevice::SD_ErrCode sderr;
    unsigned char data;
 
	sderr = sd->SD_GetADFStatus(&data);
	*adfStat = data;
	
	if (sderr != ScannerDevice::SD_OK)
		return SANE_STATUS_INVAL;
	
	return SANE_STATUS_GOOD;
}

extern "C" void llpddk_report_error(
	int error)
{
	ei->ReportError((ScanErrorInterface::ScannerError) error);
}

