/**************************************************************************
   saneportmonitor.cpp - SANE library for Lexmark scanners.
                       - reads/writes from/to bulk endpoints
   Copyright (C) 2003-2005 Lexmark International, Inc.
 
   This file is part of the SANE package.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston,
   MA 02111-1307, USA.

   As a special exception, the authors of SANE give permission for
   additional uses of the libraries contained in this release of SANE.

   The exception is that, if you link a SANE library with other files
   to produce an executable, this does not by itself cause the
   resulting executable to be covered by the GNU General Public
   License.  Your use of that executable is in no way restricted on
   account of linking the SANE library code into it.

   This exception does not, however, invalidate any other reasons why
   the executable file might be covered by the GNU General Public
   License.

   If you submit changes to SANE to the maintainers to be included in
   a subsequent release, you agree by submitting the changes that
   those changes may be distributed with this exception intact.

   If you write modifications of your own for SANE, it is your choice
   whether to permit this exception to apply to your modifications.
   If you do not wish that, delete this exception notice.
   **************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "saneportmonitor.h"

extern "C" void sanei_usb_init(
	void);
extern "C" SANE_Status sanei_usb_open (
	SANE_String_Const devname, 
	SANE_Int * dn);
extern "C" void sanei_usb_close (
	SANE_Int dn);
extern "C" SANE_Status sanei_usb_read_bulk (
	SANE_Int dn, 
	SANE_Byte * buffer, 
	size_t * size);
extern "C" SANE_Status sanei_usb_write_bulk (
	SANE_Int dn, 
	const SANE_Byte * buffer, 
	size_t * size);
extern "C" void sanei_usb_attach_matching_devices (
	const char *name,
	SANE_Status (*attach) (const char *dev));

SanePortMonitor::SanePortMonitor()
{
	// make sure port handle is in default state
	m_PortHandle = SANE_PM_INVALID_HANDLE;

	// as well as the device name
	m_DeviceName = NULL;

	// initialize sanei functions
	sanei_usb_init();
}

SanePortMonitor::~SanePortMonitor()
{
	if (m_PortHandle != SANE_PM_INVALID_HANDLE)
		PM_Close();
	
	if (m_DeviceName)
		free(m_DeviceName);
	
	m_PortHandle = SANE_PM_INVALID_HANDLE;
	m_DeviceName = NULL;
}
		
PortMonitor::PM_Error SanePortMonitor::PM_Open()
{
	SANE_Status devstatus;
	
	if (!m_DeviceName)
		return PortMonitor::PM_ERR_FAILED;
		
	devstatus = sanei_usb_open(m_DeviceName, &m_PortHandle);
	if (devstatus != SANE_STATUS_GOOD)
		return PortMonitor::PM_ERR_FAILED;

	return PortMonitor::PM_ERR_SUCCESS;
}

PortMonitor::PM_Error SanePortMonitor::PM_Close()
{
	if (m_PortHandle != SANE_PM_INVALID_HANDLE)
		sanei_usb_close(m_PortHandle);

	m_PortHandle = SANE_PM_INVALID_HANDLE;

	return PortMonitor::PM_ERR_SUCCESS;
}

long SanePortMonitor::PM_Write(
	unsigned char * buffer,
	unsigned long size)
{
	SANE_Status devstatus;
	long actualwrite;

	actualwrite = (long) size;
	
	if (m_PortHandle == SANE_PM_INVALID_HANDLE)
		return -1;

	sanei_usb_close(m_PortHandle);
	sanei_usb_init();
	devstatus = sanei_usb_open(m_DeviceName, &m_PortHandle);
	devstatus = sanei_usb_write_bulk(
		m_PortHandle,
		(SANE_Byte *) buffer,
		(size_t *) &actualwrite);

	if (devstatus != SANE_STATUS_GOOD)
		return -1;
	
	return actualwrite;
}

long SanePortMonitor::PM_Read(
	unsigned char * buffer,
	unsigned long size)
{
	SANE_Status devstatus;
	long actualread;
	actualread = (long) size;
	
	if (m_PortHandle == SANE_PM_INVALID_HANDLE)
		return -1;

	devstatus = sanei_usb_read_bulk(
		m_PortHandle,
		(SANE_Byte *) buffer,
		(size_t *) &actualread);

	if (devstatus == SANE_STATUS_EOF)
		return 0;
	
	if (devstatus != SANE_STATUS_GOOD)
		return -1;
	
	return actualread;
}

PortMonitor::PM_Error SanePortMonitor::PM_RegisterEventCallback(
	PM_EventCallback pfnCallback)
{
	// not implemented
	return PortMonitor::PM_ERR_SUCCESS;
}

PortMonitor::PM_Error SanePortMonitor::PM_UnregisterEventCallback()
{
	// not implemented
	return PortMonitor::PM_ERR_SUCCESS;
}

PortMonitor::PM_Error SanePortMonitor::PM_WaitForData(
	long timeout)
{
	// not implemented
	return PortMonitor::PM_ERR_SUCCESS;
}

PortMonitor::PM_Error SanePortMonitor::PM_SetDeviceName(
	SANE_String_Const devname)
{
	m_DeviceName = strdup((char *) devname);
	if (m_DeviceName == NULL)
		return PortMonitor::PM_ERR_FAILED;

	return PortMonitor::PM_ERR_SUCCESS;
}

PortMonitor::PM_Error SanePortMonitor::PM_AttachDevice(
	const char * name,
	SANE_Status (* attach) (const char * dev))
{
	if (attach == NULL)
		return PortMonitor::PM_ERR_FAILED;
	
	sanei_usb_attach_matching_devices(
		name, attach);	

	return PortMonitor::PM_ERR_SUCCESS;
}

